@@ -87,6 +87,8 @@ MARK_VISIBLE (_gpgrt_get_std_stream)
 MARK_VISIBLE (gpgrt_flockfile)
 MARK_VISIBLE (gpgrt_ftrylockfile)
 MARK_VISIBLE (gpgrt_funlockfile)
+MARK_VISIBLE (_gpgrt_pending)
+MARK_VISIBLE (_gpgrt_pending_unlocked)
 MARK_VISIBLE (gpgrt_feof)
 MARK_VISIBLE (gpgrt_feof_unlocked)
 MARK_VISIBLE (gpgrt_ferror)
@@ -125,6 +127,9 @@ MARK_VISIBLE (gpgrt_vfprintf_unlocked)
 MARK_VISIBLE (gpgrt_setvbuf)
 MARK_VISIBLE (gpgrt_setbuf)
 MARK_VISIBLE (gpgrt_set_binary)
+MARK_VISIBLE (gpgrt_set_nonblock)
+MARK_VISIBLE (gpgrt_get_nonblock)
+MARK_VISIBLE (gpgrt_poll)
 MARK_VISIBLE (gpgrt_tmpfile)
 MARK_VISIBLE (gpgrt_opaque_set)
 MARK_VISIBLE (gpgrt_opaque_get)
@@ -190,6 +195,8 @@ MARK_VISIBLE (gpgrt_set_alloc_func)
 #define gpgrt_flockfile             _gpgrt_USE_UNDERSCORED_FUNCTION
 #define gpgrt_ftrylockfile          _gpgrt_USE_UNDERSCORED_FUNCTION
 #define gpgrt_funlockfile           _gpgrt_USE_UNDERSCORED_FUNCTION
+#define _gpgrt_pending              _gpgrt_USE_UNDERSCORED_FUNCTION
+#define _gpgrt_pending_unlocked     _gpgrt_USE_UNDERSCORED_FUNCTION
 #define gpgrt_feof                  _gpgrt_USE_UNDERSCORED_FUNCTION
 #define gpgrt_feof_unlocked         _gpgrt_USE_UNDERSCORED_FUNCTION
 #define gpgrt_ferror                _gpgrt_USE_UNDERSCORED_FUNCTION
@@ -228,6 +235,9 @@ MARK_VISIBLE (gpgrt_set_alloc_func)
 #define gpgrt_setvbuf               _gpgrt_USE_UNDERSCORED_FUNCTION
 #define gpgrt_setbuf                _gpgrt_USE_UNDERSCORED_FUNCTION
 #define gpgrt_set_binary            _gpgrt_USE_UNDERSCORED_FUNCTION
+#define gpgrt_set_nonblock          _gpgrt_USE_UNDERSCORED_FUNCTION
+#define gpgrt_get_nonblock          _gpgrt_USE_UNDERSCORED_FUNCTION
+#define gpgrt_poll                  _gpgrt_USE_UNDERSCORED_FUNCTION
 #define gpgrt_tmpfile               _gpgrt_USE_UNDERSCORED_FUNCTION
 #define gpgrt_opaque_set            _gpgrt_USE_UNDERSCORED_FUNCTION
 #define gpgrt_opaque_get            _gpgrt_USE_UNDERSCORED_FUNCTION
